;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;+         
;          To determine the relative positions of four
;	   user-selected line crossings of a target image. 
;          Relative Sub-pixel coordinates are obtained by cross-correlating
;          small squares centered on cursor-selected locations which should
;          contain one line crossing per square. The un-binned image must have
;          been displayed on the ACTIVE graphic window BEFORE calling this
;          routine! It is recommended to display both, continuum and filter 
;          image before calling XY_4PNTS_CORR.
; CALLING SEQUENCE:
;	   xy4 = XY_4PNTS_CORR (image_array,name [,XOFF=xoff] ...
;                ... [,KERNEL_SIZE=kernsz] [,FLAG=flag] )
; INPUTS:
;	   image_array: target image of  continuum or filter chanel.
;          name:        string, e.g.: 'continuum' or 'filter' (for print-out 
;                       only).
; OPTIONAL INPUT PARAMETER:
;	   XOFF=xoff: horizontal pixel-offset of the displayed image on the
;                     graphic window; default: xoff = 0.
;          KERNEL_SIZE=kernsz: size of the four squares used for correlation;
;                     default: kernsz=10
; OUTPUTS:
;	   xy4 : fltarr(4,2), the pixel-coordinates (incl. fractional part)
;                of the centres of the four boxes after being shifted such
;                that the correlation with the 1st box is maximal;
;                xy4(i,0), xy4(i,1) are the corrected x- and y-coordinates
;                for box <i>.
; OPTIONAL OUTPUTS:
;	   FLAG=flag: = 1  : no problems encountered; 
;                     = -1 : severe problem, no cross-correlation for at least
;                           one pair of boxes; returned coordinates unreliable.
;                     > 1  : correlation coefficient seems to be too low in
;                            flag -1 cases, returned coordinates uncertain.
;
;; 2000-Jun-28, nlte@kis: created.
;; 2011-Jun-22, ali@nso: modified.
;;
;;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
FUNCTION ibis_xy_4pnts_corr,image,name,XOFF=xoff,KERNEL_SIZE=kernsz,FLAG=flag,$
                       HELP=help

on_error,1
;
if n_params() ne 2 or keyword_set(help) then begin
   doc_library,'xy_4pnts_corr'
   return,-1
endif
;
if n_elements(xoff) eq 0 then xoff=0
if n_elements(kernsz) eq 0 then kernsz=10  ; default for correlation box size
if n_elements(dbad) eq 0 then dnogood=5 else dnogood=dbad 
; bad position measurement: deviation >= dnogood pixel ; default = 5
;
cpoint=['first ','second','third ','fourth']
szx=(size(image))(1) & szy=(size(image))(2)
k2=(kernsz+1)/2
fbin=3
;
x4=intarr(4) & y4=intarr(4)
flag4=0
;
; let user click four positions of target
if strpos(name,'Cont') ge 0 then $
   print,name+' image: select 4 good target points!' else $
   print,name+' image: identify the same 4 target points!'
   print,name+' go clockwise with 1st point in upper left corner!'
for ip=0,3 do begin
    print,' Move cursor to '+cpoint(ip)+$
          ' selected cross and klick left mouse button.'+string(7b) ; with beep
     CURSOR,xc,yc,3,/DEV
     x4(ip)=xc-xoff & y4(ip)=yc
     plots,/dev,[xc-k2,xc-k2,xc+k2,xc+k2,xc-k2],$
                [yc-k2,yc+k2,yc+k2,yc-k2,yc-k2]
     xyouts,/dev,xc+k2+2,yc,string(ip+1,form='(i0)'),chars=1.5
endfor  ; 4 points
;
; find shifts for maximal correlation of the subimages centered on
; these 4 points (2,3,4 with 1):
xycmean=fltarr(4,2)
nbin=fbin*(2*k2+1)
maxshift=nint(float(2*k2+1)/3.)*fbin
w_unshmask=(k2/2)>3
;
xc=x4(0) & yc=y4(0)
xycmean(0,0)=xc & xycmean(0,1)=yc 
ref=rebin(UNSHMASK(float(image(xc-k2:xc+k2 ,yc-k2:yc+k2)),WINDOW=w_unshmask,$
          /MEDIAN),nbin,nbin,/SAMPLE)
;
;test1-start
;print,name+' ref box xc,yc,k2,fbin,w_unshm:',xc,yc,k2,fbin,w_unshmask
;tbin=(2*maxshift+1)*5 ; test
;iwact=!d.window
;if strpos(name,'Cont') ge 0 then iw=0 else iw=1
;window,iw,xs=512,ys=4*(tbin+5)
;tvbs,cr=[6,255],rebin(image(xc-k2:xc+k2 ,yc-k2:yc+k2),nbin,nbin,/SAMPLE)
;tvbs,cr=[6,255],ref,off=[nbin+5,0]
;test1-end

for ip=0,3 do begin
  if ip eq 0 then begin 
     dx=0. & dy=0. & maxcc=1. & flag=1 & txt='box = ref'
  endif else begin
     xc=x4(ip) & yc=y4(ip)  
     box=rebin(UNSHMASK(float(image(xc-k2:xc+k2 ,yc-k2:yc+k2)),$
               WINDOW=w_unshmask,/MED),nbin,nbin,/SAMPLE)
;test2-start
;print,name+string(ip,form='(1x,i0)')+' box xc,yc,k2,fbin:',xc,yc,k2,fbin
;tvbs,cr=[6,255],rebin(image(xc-k2:xc+k2 ,yc-k2:yc+k2),nbin,nbin,/SAMPLE),$
;     off=[0,ip*(tbin+5)]
;tvbs,cr=[6,255],box,off=[nbin+5,ip*(tbin+5)]
;test2-end
;
     IMAGE_DISPLACE,ref,box,SHIFT=maxshift+[0,0],dx,dy,MAXCC=maxcc,FLAG=flag,$
                    CC=cc
;test3-start
;print,'size box: ',(size(box))(1),(size(box))(2),' maxshift=',maxshift
;tvbs,cr=[6,255],rebin(cc,tbin,tbin,/sample),off=[2*(nbin+5)+5,ip*(tbin+5)]
;print,'from IMAGE_DISPLACE: dx,dy,flag',dx,dy,flag
;test3-end
;
     case 1 of
       flag lt 1: begin
               txt='!!! No correlation'
               dx=0. & dy=0.
               end
       flag eq 1: txt='OK'
       flag eq 2: txt='?OK?'
       else:      txt='!!! Uncertain'
     endcase 
;
     dx=dx/float(fbin) & dy=dy/float(fbin)
  endelse 
  xycmean(ip,0)=x4(ip) -dx
  xycmean(ip,1)=y4(ip) -dy
  print,ip+1,dx,dy,maxcc,flag,txt,form=$
  '("point ",i1," corrected dx,dy =",2f7.3," max. correl=",f5.2," flag= ",i2,1x,a)'
if flag lt 1 then flag4=-1
if flag gt 2 and flag4 ne -1 then flag4=flag4+1
endfor
case 1 of
   flag4 eq -1: flag=-1
   flag4 eq 0 : flag=1
   flag4 gt 0 : flag=1+flag4
endcase
;
;wset,iwact ; test
;
return,xycmean
end
