FUNCTION UNSHMASK ,imgorig,window=w0,ratio=r0,median=med
;+
; NAME:
;	UNSHMASK
; PURPOSE:
;	Returns contrast enhanced image(s) using the unsharp masking method.
; CATEGORY:            @CAT-# 16 12 
;	Image Processing , Filtering
; CALLING SEQUENCE:
;	new_image = UNSHMASK (orig_image [,WINDOW=w] [,RATIO=r] [,/MEDIAN]
; INPUTS:
;	orig_image:  2- or 3-dim array containing one or more images for
;                    which contrast enhancement shall be done. 
;                    If orig_image is a 3-dim array, the first 2 dimensions 
;                    of the array must correspond to the image format (nx,ny),
;                    the size of the 3rd dim. is then the number of images.
; OPTIONAL INPUT PARAMETER:
;	WINDOW=w  :  w is the pixel-size of the smoothing window (smoothing is
;                    done with IDL routine SMOOTH or MEDIAN, using a
;                    w*w-box; w should be an odd number; 
;                    default: w = (n/50) > 3 , n the smaller of the two image 
;                                                dimensions.
;       RATIO=r   :  r is the ratio of smoothed image to original image; if
;                    sm is the smoothed image, then 
;                         ( orig_image - r * sm ) / (1.-r)
;                    will be returned as contrast enhanced image.
;                    Default: r = 0.8
;       /MEDIAN   :  If set, median filtering (IDL MEDIAN) will be used to 
;                    instead of boxcar averaging (IDL SMOOTH) to create the 
;                    smoothed image sm. The advantage of median filtering is
;                    that large-scale edges (e.g. the solar limb of a full 
;                    disk image) will not be blurred.Median filtering is,
;                    however, much slower than boxcar averaging.
;                    
; OUTPUTS:
;	new_image :  the array containing the contrast enhanced image(s),
;                    same type and size as orig_image.
; COMMON BLOCKS:
;	none
; SIDE EFFECTS:
;	none
; RESTRICTIONS:
;	The image-size should be at least 3 im both spatial dimensions;
;       Byte-arrays should contain only non-negative values.
; PROCEDURE:
;	calls IDL-routine SMOOTH with /EDGE_TRUNCATE set or, if /MEDIAN is
;       set, IDL-routine MEDIAN, to compute the smoothed image sm and returns
;         ( orig_image - r * sm ) / (1.-r)  , where sm is the smoothed image.
;
; MODIFICATION HISTORY:
;	nlte (KIS), 1996-05-19 created
;       nlte (KIS), 1996-06-27 option median filtering
;-
on_error,2
;
if n_params() ne 1 then message,$
   'Usage: UNSHMASK (orig_image [,WINDOW=w] [,RATIO=r]'
if (size(imgorig))(0) ne 2 and (size(imgorig))(0) ne 3 then $
       message,'Argument imgorig not 2- or 3-dim' 
sz=size(imgorig)
sx=sz(1) & sy=sz(2)
if sx lt 3 or sy lt 3 then message,'Image size too small either in x or y.'
imgtype=sz(sz(0)+1)
if imgtype eq 0 or imgtype eq 7 or imgtype eq 8 then message,$
   'Argument imgorig must be of numeric type.'
imed=keyword_set(med)
; 
; set defaults for w, r :
if n_elements(w0) eq 0 then w= nint(float(min([sx,sy]))/50.) > 3 $
                       else w=nint(w0)
if (w/2)*2 eq w then w=w+1   ; makes w odd
;
if n_elements(r0) eq 0 then r=0.8 $
else begin
   if r0 le 0. or r0 ge 1. then message,'RATIO-value must be > 0, < 1.'
   r=r0
endelse
norm=1./(1.-r)
;
; 1 or more images in imgorig ?
single=(size(imgorig))(0) eq 2 
;
; type conversion neccessary ?
fpconv=0
if imed then begin
  if imgtype ne 1 and imgtype ne 4 then fpconv=1
endif else begin
if imgtype le 2 then fpconv=1 else if imgtype eq 3 then fpconv=2
endelse
;
if not single then begin            ; imgorig contains nimg images
  nimg=sz(3)
  contr_img = imgorig
  for i=0,nimg-1 do begin
    case fpconv of
        1 : img=float(reform(imgorig(*,*,i)))
        2 : img=double(reform(imgorig(*,*,i)))
      else: img=reform(imgorig(*,*,i))
    endcase
    if imed then sm=MEDIAN(img,w) else sm=SMOOTH(img,w,/edge_trunc)
    img=(temporary(img)-r*sm)*norm
    case imgtype of
        1  : contr_img(*,*,i) = byte( nint(img) > 0 )
        2  : contr_img(*,*,i) = nint(img)
        3  : contr_img(*,*,i) = nint(img,/long)
      else : contr_img(*,*,i) = img
    endcase
  endfor
;
endif else begin                    ; imgorig is 2-dim, contains 1 image 
  case fpconv of
     1  : img=float(imgorig)
     2  : img=double(imgorig)
    else: img=imgorig
  endcase
  if imed then sm=MEDIAN(img,w) else sm=SMOOTH(img,w,/edge_trunc)
  img=(temporary(img)-r*sm)*norm
  case imgtype of
      1  : contr_img = byte( nint(img) > 0 )
      2  : contr_img = nint(img)
      3  : contr_img = nint(img,/long)
    else : contr_img = img
  endcase
;
endelse
;
jmpret:
sm=0 & img=0
return,contr_img
;
end
